/**
 * Author: Heri, Alex, Dan
 * Date: 31mar15
 * Ver: 2
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <libspe2.h>
#include <pthread.h>

#define SPU_THREADS 16

extern spe_program_handle_t spu_mailbox;

typedef struct {
   long cellno;
   int data;
   spe_context_ptr_t spe;
} thread_arg_t;

void* ppu_pthread_function(void* thread_arg)
{
   thread_arg_t* arg = (thread_arg_t*)thread_arg;
   unsigned int entry = SPE_DEFAULT_ENTRY;

   if(spe_context_run(arg->spe, &entry, 0, (void*)arg->cellno,
      (void*)(long)arg->data, NULL) < 0) {
      perror("Failed running context");
      exit(1);
   }

   pthread_exit(NULL);
}

int main(void)
{
   int i;
   spe_event_handler_ptr_t event_handler;
   spe_event_unit_t events[SPU_THREADS];
   spe_context_ptr_t ctxs[SPU_THREADS];
   pthread_t threads[SPU_THREADS];
   thread_arg_t arg[SPU_THREADS];

   /* Create event handler for working with events. */
   if((event_handler = spe_event_handler_create()) == NULL) {
      perror("Failed creating event handler");
      exit(1);
   }

   /* Create several event-enabled SPE-threads to execute 'SPU'. */
   for(i=0; i<SPU_THREADS; i++) {
      /* Create context */
      if((ctxs[i] = spe_context_create(SPE_EVENTS_ENABLE, NULL)) == NULL) {
         perror("Failed creating context");
         exit(1);
      }

      /* Load program into context */
      if(spe_program_load(ctxs[i], &spu_mailbox)) {
         perror("Failed loading program");
         exit(1);
      }

      /* Send some more paramters to the SPU. */
      arg[i].cellno = i;
      arg[i].spe = ctxs[i];
      arg[i].data = 55 + i;

      /* Create thread for each SPE context */
      if(pthread_create(&threads[i], NULL, &ppu_pthread_function, (void*)&arg[i])) {
         perror("Failed creating thread");
         exit(1);
      }

      /* Register event for message available in the out_intr mailbox of this SPU. */
      events[i].events = SPE_EVENT_OUT_INTR_MBOX;
      events[i].spe = ctxs[i];
      events[i].data.ptr = (void*)&arg[i]; // recover this thread's info when event arrives
      if(spe_event_handler_register(event_handler, &events[i]) < 0) {
         perror("Failed event registration.");
         exit(1);
      }
   }

   /* Wait for data and print it. */
   /* Optimal; we receive in the sent order. */
   for(i=0; i<SPU_THREADS; i++) {
      int ret;
      spe_event_unit_t event;
      unsigned int mbox_data;
      thread_arg_t* arg;

      /* Wait for event. */
      if((ret = spe_event_wait(event_handler, &event, 1, -1)) < 0) {
         perror("Failed event_wait.");
         exit(1);
      }

      /* event will contain info about SPU that caused the event. */
      arg = (thread_arg_t*)event.data.ptr;

      /* Read data; will not block because we had an event on this SPU and mailbox. */
      spe_out_intr_mbox_read(arg->spe, &mbox_data, 1, SPE_MBOX_ALL_BLOCKING);
      printf("[PPU] SPU %ld sent data = %d\n", arg->cellno, (int)mbox_data);

      if((int)mbox_data == arg->data)
           printf("[PPU] SPU %ld data checks\n", arg->cellno);
      else
           printf("[PPU] SPU %ld data check failed\n", arg->cellno);
   }

   /* Wait for SPU-thread to complete execution. */
   for(i=0; i<SPU_THREADS; i++) {
      if(pthread_join(threads[i], NULL)) {
         perror("Failed pthread_join");
         exit(1);
      }

      /* Deregister event */
      if(spe_event_handler_deregister(event_handler, &events[i]) < 0) {
         perror("Failed event deregister.");
         exit(1);
      }

      /* Destroy context */
      if(spe_context_destroy(ctxs[i]) != 0) {
         perror("Failed destroying context");
         exit(1);
      }
   }

   /* Destory event handler */
   if(spe_event_handler_destroy(event_handler) < 0) {
      perror("Failed destroying event handler.");
      exit(1);
   }

   printf("\nThe program has successfully executed.\n");
   return 0;
}

